#include "Fixed_Variable_Channel.H"

#include "Channel_Elements.H"

using namespace PHASIC;

Fixed_Variable_Channel::
Fixed_Variable_Channel(int nin,int nout,ATOOLS::Flavour *fl,
		       const std::string &variable):
  PHASIC::Channel_Interface(nin,nout,fl),
  p_variable(ATOOLS::Variable_Getter::GetObject(variable,"")) {}

void Fixed_Variable_Channel::
GeneratePoint(ATOOLS::Vec4D *p,double *ran)
{
  switch (p_variable->SelectorID()) {
  case 12: {
    Ehat=sqrt((p[0]+p[1]).Abs2());
    pt=m_value;
    if (Ehat/2.0>pt) {
      weight=1.0/sqrt((Ehat*Ehat)/(4.*pt*pt)-1.0);
      p[2]=ATOOLS::Vec4D(Ehat/2.0,pt*cos(2.0*M_PI*ran[1]),
			 pt*sin(2.0*M_PI*ran[1]),
			 sqrt(Ehat*Ehat/4.0-pt*pt));
      m_trigger=true;
    }
    else {
      weight=ATOOLS::Accu();
      m_trigger=false;
      p[2]=ATOOLS::Vec4D(Ehat/2.0,Ehat/2.0*cos(2.0*M_PI*ran[1]),
			 Ehat/2.0*sin(2.0*M_PI*ran[1]),0.0);
    }
    p[3]=ATOOLS::Vec4D(Ehat/2.0,ATOOLS::Vec3D()-ATOOLS::Vec3D(p[2]));
    break;
  }
  default:
    msg_Error()<<"Fixed_Variable_Channel::GeneratePoint(..): "
		       <<"Cannot handle "<<p_variable->Name()
		       <<"! Setting weight to 0."<<std::endl;
    weight=0.0;
  }
}
  
void Fixed_Variable_Channel::GenerateWeight(ATOOLS::Vec4D *_p)
{
  weight/=PHASIC::CE.Isotropic2Weight(_p[2],_p[3])*
    pow(2.0*M_PI,2.0*3.0-4.0);
}

std::string Fixed_Variable_Channel::ChID()
{
  return m_chid;
}
