#include "Standard_Model.H"
#include "Running_AlphaQED.H"
#include "Running_AlphaS.H"
#include "Running_Fermion_Mass.H"
#include "Message.H"
#include "Data_Reader.H"
#include <iomanip>

using namespace MODEL;
using namespace ATOOLS;

DECLARE_GETTER(Standard_Model_Getter,"SM",Model_Base,Model_Arguments);

Model_Base *Standard_Model_Getter::operator()(const Model_Arguments &args) const
{
  return new Standard_Model(args.m_path,args.m_file,args.m_elementary);
}

void Standard_Model_Getter::PrintInfo(std::ostream &str,const size_t width) const
{ 
  str<<"The Standard Model\n";
  str<<std::setw(width+4)<<" "<<"{\n"
     <<std::setw(width+7)<<" "<<"parameter specification [keyword=value]\n"
     <<std::setw(width+7)<<" "<<"- EW_SCHEME (values 0,1,2,3, EW input schemes, see documentation)\n"
     <<std::setw(width+7)<<" "<<"- ALPHAS(MZ) (strong coupling at MZ)\n"
     <<std::setw(width+7)<<" "<<"- ALPHAS(default) (fixed strong coupling)\n"
     <<std::setw(width+7)<<" "<<"- ORDER_ALPHAS (0,1,2 -> 1, 2, 3-loop running)\n"
     <<std::setw(width+7)<<" "<<"- 1/ALPHAQED(0) (alpha QED Thompson limit)\n"
     <<std::setw(width+7)<<" "<<"- 1/ALPHAQED(default) (fixed alpha QED)\n"
     <<std::setw(width+7)<<" "<<"- SIN2THETAW (weak mixing angle)\n"
     <<std::setw(width+7)<<" "<<"- VEV (Higgs vev)\n"
     <<std::setw(width+7)<<" "<<"- LAMBDA (Higgs quartic coupling)\n"
     <<std::setw(width+7)<<" "<<"- CKMORDER (0,1,2,3 - order of CKM expansion in Cabibbo angle)\n"
     <<std::setw(width+7)<<" "<<"- CABIBBO (Cabibbo angle in Wolfenstein parameterization)\n"
     <<std::setw(width+7)<<" "<<"- A (Wolfenstein A)\n"
     <<std::setw(width+7)<<" "<<"- RHO (Wolfenstein Rho)\n"
     <<std::setw(width+7)<<" "<<"- ETA (Wolfenstein Eta)\n"
     <<std::setw(width+4)<<" "<<"}";
}

DECLARE_GETTER(Standard_Model_Top_Getter,"SM+Top",Model_Base,Model_Arguments);

Model_Base *Standard_Model_Top_Getter::operator()(const Model_Arguments &args) const
{
  return new Standard_Model(args.m_path,args.m_file,args.m_elementary,1);
}

void Standard_Model_Top_Getter::PrintInfo(std::ostream &str,const size_t width) const
{ 
  str<<"The Standard Model + non-standard top couplings\n";
  str<<std::setw(width+4)<<" "<<"{\n"
     <<std::setw(width+7)<<" "<<"parameter specification [keyword=value]\n"
     <<std::setw(width+7)<<" "<<"- EW_SCHEME (values 0,1,2,3, EW input schemes, see documentation)\n"
     <<std::setw(width+7)<<" "<<"- ALPHAS(MZ) (strong coupling at MZ)\n"
     <<std::setw(width+7)<<" "<<"- ALPHAS(default) (fixed strong coupling)\n"
     <<std::setw(width+7)<<" "<<"- ORDER_ALPHAS (0,1,2 -> 1, 2, 3-loop running)\n"
     <<std::setw(width+7)<<" "<<"- 1/ALPHAQED(0) (alpha QED Thompson limit)\n"
     <<std::setw(width+7)<<" "<<"- 1/ALPHAQED(default) (fixed alpha QED)\n"
     <<std::setw(width+7)<<" "<<"- SIN2THETAW (weak mixing angle)\n"
     <<std::setw(width+7)<<" "<<"- VEV (Higgs vev)\n"
     <<std::setw(width+7)<<" "<<"- LAMBDA (Higgs quartic coupling)\n"
     <<std::setw(width+7)<<" "<<"- CKMORDER (0,1,2,3 - order of CKM expansion in Cabibbo angle)\n"
     <<std::setw(width+7)<<" "<<"- CABIBBO (Cabibbo angle in Wolfenstein parameterization)\n"
     <<std::setw(width+7)<<" "<<"- A (Wolfenstein A)\n"
     <<std::setw(width+7)<<" "<<"- RHO (Wolfenstein Rho)\n"
     <<std::setw(width+7)<<" "<<"- ETA (Wolfenstein Eta)\n"
     <<std::setw(width+7)<<" "<<"- KAPPA_{TBW} (rel. factor of tbW w.r.t. SM)\n"
     <<std::setw(width+7)<<" "<<"- THETA_{TBW} (angle of tbW in PL-PR w.r.t. SM)\n"
     <<std::setw(width+4)<<" "<<"}";
}


Standard_Model::Standard_Model(std::string _dir,std::string _file,
			       bool _elementary,int _trivialextension) :
  Model_Base(_dir,_file,_elementary), m_trivialextension(_trivialextension)
{
  if (m_elementary) 
    msg_Info()<<"Initialize the Standard Model from "<<m_dir<<" / "<<m_file<<std::endl;
  m_name      = std::string("SM");
  p_numbers          = new ScalarNumbersMap();
  p_constants        = new ScalarConstantsMap();
  p_complexconstants = new ComplexConstantsMap();
  p_functions        = new ScalarFunctionsMap();
  p_matrices         = new ComplexMatricesMap();
  
  ParticleInit();
  FillSpectrum();
  if (m_elementary) {
    ATOOLS::OutputParticles(msg->Info());
    ATOOLS::OutputContainers(msg->Info());
  }
}

void Standard_Model::ParticleInit() {
  //add SM particles
  //kf_code,mass,width,charge,icharge,strong,spin,majorana,take,stable,massive,idname,tex_name
  s_kftable[1] = new  Particle_Info(1,0.01,.0,-1,-1,3,1,0,1,1,0,"d","d");
  s_kftable[2] = new  Particle_Info(2,0.005,.0,2,1,3,1,0,1,1,0,"u","u");
  s_kftable[3] = new  Particle_Info(3,0.2,.0,-1,-1,3,1,0,1,1,0,"s","s");
  s_kftable[4] = new  Particle_Info(4,1.42,.0,2,1,3,1,0,1,1,0,"c","c");
  s_kftable[5] = new  Particle_Info(5,4.8,.0,-1,-1,3,1,0,1,1,0,"b","b");
  s_kftable[6] = new  Particle_Info(6,175.,1.5,2,1,3,1,0,1,1,1,"t","t");
  s_kftable[11] = new Particle_Info(11,0.000511,.0,-3,-1,0,1,0,1,1,0,"e-","e^-");
  s_kftable[12] = new Particle_Info(12,.0,.0,0,1,0,1,0,1,1,0,"nu_e","\\nu_e");
  s_kftable[13] = new Particle_Info(13,.105,.0,-3,-1,0,1,0,1,1,0,"mu-","\\mu^-");
  s_kftable[14] = new Particle_Info(14,.0,.0,0,1,0,1,0,1,1,0,"nu_mu","\\nu_\\mu");
  s_kftable[15] = new Particle_Info(15,1.777,2.36E-12,-3,-1,0,1,0,1,1,0,"tau-","\\tau^-");
  s_kftable[16] = new Particle_Info(16,.0,.0,0,1,0,1,0,1,1,0,"nu_tau","\\nu_\\tau");
  s_kftable[21] = new Particle_Info(21,.0,.0,0,0,8,2,-1,1,1,0,"G","g");
  s_kftable[22] = new Particle_Info(22,.0,.0,0,0,0,2,-1,1,1,0,"P","\\gamma");
  s_kftable[23] = new Particle_Info(23,91.188,2.49,0,0,0,2,-1,1,0,1,"Z","Z");
  s_kftable[24] = new Particle_Info(24,80.419,2.06,3,0,0,2,0,1,0,1,"W+","W^+");
  s_kftable[25] = new Particle_Info(25,120.,0.0037,0,0,0,0,-1,1,0,1,"h0","h_0");

  ReadParticleData();

  // add pseudo particles and containers
  s_kftable[kf_none] = new
    Particle_Info(kf_none,-1,0,0,0,0,0,-1,0,1,0,"no_particle","no particle");
  s_kftable[kf_seaquark] = new
    Particle_Info(kf_seaquark,0.,0.,0,0,1,1,0,1,1,0,"seaquark","seaquark");
  s_kftable[kf_bjet] = new
    Particle_Info(kf_bjet,0.,0.,0,0,1,2,0,1,1,0,"bj","bjet");
  s_kftable[kf_cjet] = new
    Particle_Info(kf_cjet,0.,0.,0,0,1,2,0,1,1,0,"cj","cjet");

  s_kftable[kf_fermion] = new
    Particle_Info(kf_fermion,0.,0., 0,0,0,1,0,1,1,0,"fermion","fermion");
  s_kftable[kf_jet] = new
    Particle_Info(kf_jet,0.,0.,0,0,1, 2,0,1,1,0,"j","jet");
  s_kftable[kf_quark] = new
    Particle_Info(kf_quark,0.,0.,0, 0,1,1,0,1,1,0,"Q","Quark");
  s_kftable[kf_lepton] = new
    Particle_Info(kf_lepton,0.,0.,-3,-1,0,1,0,1,1,0,"lepton","lepton");
  s_kftable[kf_neutrino] = new
    Particle_Info(kf_neutrino,0.,0.,0,1,0, 1,0,1,1,0,"neutrino","neutrino");
  s_kftable[kf_fermion]->Clear();
  s_kftable[kf_jet]->Clear();
  s_kftable[kf_quark]->Clear();
  s_kftable[kf_lepton]->Clear();
  s_kftable[kf_neutrino]->Clear();
  for (int i=1;i<7;i++) {
    Flavour addit((kf_code)i);
    if (addit.Mass()==0.0 || !addit.IsMassive()) {
      s_kftable[kf_jet]->Add(addit);
      s_kftable[kf_jet]->Add(addit.Bar());
      s_kftable[kf_quark]->Add(addit);
      s_kftable[kf_quark]->Add(addit.Bar());
      s_kftable[kf_fermion]->Add(addit);
      s_kftable[kf_fermion]->Add(addit.Bar());
    }
  }
  s_kftable[kf_jet]->Add(Flavour(kf_gluon));
  for (int i=11;i<17;i+=2) {
    Flavour addit((kf_code)i);
    if (addit.Mass()==0.0 || !addit.IsMassive()) {
      s_kftable[kf_lepton]->Add(addit);
      s_kftable[kf_lepton]->Add(addit.Bar());
      s_kftable[kf_fermion]->Add(addit);
      s_kftable[kf_fermion]->Add(addit.Bar());
    }
  }
  for (int i=12;i<18;i+=2) {
    Flavour addit((kf_code)i);
    if (addit.Mass()==0.0) {
      s_kftable[kf_neutrino]->Add(addit);
      s_kftable[kf_neutrino]->Add(addit.Bar());
      s_kftable[kf_fermion]->Add(addit);
      s_kftable[kf_fermion]->Add(addit.Bar());
    }
  }
}

void Standard_Model::FillSpectrum() {
  p_dataread = new Data_Reader(" ",";","!","=");
  p_dataread->AddWordSeparator("\t");
  p_dataread->SetInputPath(m_dir);
  p_dataread->SetInputFile(m_file);
  FixEWParameters();  
  FixCKM();
  p_constants->insert(std::make_pair(std::string("Yukawa_e"), 
				     p_dataread->GetValue<double>("YUKAWA_E",0.)));
  p_constants->insert(std::make_pair(std::string("Yukawa_mu"), 
				     p_dataread->GetValue<double>("YUKAWA_MU",0.)));
  p_constants->insert(std::make_pair(std::string("Yukawa_tau"), 
				     p_dataread->GetValue<double>("YUKAWA_TAU",Flavour(kf_tau).PSMass())));
  p_constants->insert(std::make_pair(std::string("Yukawa_d"), 
				     p_dataread->GetValue<double>("YUKAWA_D",0.)));
  p_constants->insert(std::make_pair(std::string("Yukawa_u"), 
				     p_dataread->GetValue<double>("YUKAWA_U",0.)));
  p_constants->insert(std::make_pair(std::string("Yukawa_s"), 
				     p_dataread->GetValue<double>("YUKAWA_S",0.)));
  p_constants->insert(std::make_pair(std::string("Yukawa_c"), 
				     p_dataread->GetValue<double>("YUKAWA_C",0.)));
  p_constants->insert(std::make_pair(std::string("Yukawa_b"), 
				     p_dataread->GetValue<double>("YUKAWA_B",Flavour(kf_b).PSMass())));
  p_constants->insert(std::make_pair(std::string("Yukawa_t"), 
				     p_dataread->GetValue<double>("YUKAWA_T",Flavour(kf_t).PSMass())));
  // Extra coupling parameters for non-Standard tbW coupling
  p_constants->insert(std::make_pair(std::string("tbW_RelFactor"),
				     m_trivialextension==1?
				     p_dataread->GetValue<double>("KAPPA_{TBW}",1.):1.));
  p_constants->insert(std::make_pair(std::string("tbW_Angle"),
				     m_trivialextension==1?
				     p_dataread->GetValue<double>("THETA_{TBW}",0.):0.));

  int    order_alphaS	= p_dataread->GetValue<int>("ORDER_ALPHAS",1);
  double alphaS         = p_dataread->GetValue<double>("ALPHAS(MZ)",0.118);
  double alphaS_default = p_dataread->GetValue<double>("ALPHAS(default)",alphaS);
  double MZ2            = sqr((*p_constants)[std::string("MZ")]);

  as = new Running_AlphaS(alphaS,MZ2,order_alphaS);
  as->SetDefault(alphaS_default);

  p_constants->insert(std::make_pair(std::string("alpha_S(MZ)"),alphaS));
  p_functions->insert(std::make_pair(std::string("alpha_S"),as));

  Running_Fermion_Mass * md   = new Running_Fermion_Mass(Flavour(kf_d),
							 ScalarConstant(std::string("Yukawa_d")),as);
  Running_Fermion_Mass * mu   = new Running_Fermion_Mass(Flavour(kf_u),
							 ScalarConstant(std::string("Yukawa_u")),as);
  Running_Fermion_Mass * ms   = new Running_Fermion_Mass(Flavour(kf_s),
							 ScalarConstant(std::string("Yukawa_s")),as);
  Running_Fermion_Mass * mc   = new Running_Fermion_Mass(Flavour(kf_c),
							 ScalarConstant(std::string("Yukawa_c")),as);
  Running_Fermion_Mass * mb   = new Running_Fermion_Mass(Flavour(kf_b),
							 ScalarConstant(std::string("Yukawa_b")),as);
  Running_Fermion_Mass * mt   = new Running_Fermion_Mass(Flavour(kf_t),
							 ScalarConstant(std::string("Yukawa_t")),as);
  Running_Fermion_Mass * me   = new Running_Fermion_Mass(Flavour(kf_e),
							 ScalarConstant(std::string("Yukawa_e")),as);
  Running_Fermion_Mass * mmu  = new Running_Fermion_Mass(Flavour(kf_mu),
							 ScalarConstant(std::string("Yukawa_mu")),as);
  Running_Fermion_Mass * mtau = new Running_Fermion_Mass(Flavour(kf_tau),
							 ScalarConstant(std::string("Yukawa_tau")),as);
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_d).IDName()),md));
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_u).IDName()),mu));
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_s).IDName()),ms));
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_c).IDName()),mc));
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_b).IDName()),mb));
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_t).IDName()),mt));
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_e).IDName()),me));
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_mu).IDName()),mmu));
  p_functions->insert(std::make_pair(std::string("m")+std::string(Flavour(kf_tau).IDName()),mtau));
  
}

void Standard_Model::FixEWParameters() {
  double MW,MZ,MH,alphaQED,sin2thetaW,cos2thetaW,vev,lambdaH,GF=1.16639e-5;
  m_ewscheme = p_dataread->GetValue<int>("EW_SCHEME",0);
  switch (m_ewscheme) {
  case 1:
    // SM parameters given by alphaQED, M_W, M_Z, M_H
    alphaQED   = 1./p_dataread->GetValue<double>("1/ALPHAQED(0)",137.03599976);
    MW         = Flavour(kf_Wplus).Mass();
    MZ         = Flavour(kf_Z).Mass();
    MH         = Flavour(kf_h0).Mass();
    cos2thetaW = sqr(MW/MZ);
    sin2thetaW = 1.-cos2thetaW;
    vev        = 2.*MW*sqrt(sin2thetaW/(4.*M_PI*alphaQED));
    lambdaH    = 2.*sqr(MH/vev);
    break;
  case 2:
    // SM parameters given by alphaQED, sinthetaW, v, M_H
    alphaQED   = 1./p_dataread->GetValue<double>("1/ALPHAQED(0)",137.03599976);
    vev        = p_dataread->GetValue<double>("VEV",246.);
    sin2thetaW = p_dataread->GetValue<double>("SIN2THETAW",0.23);
    cos2thetaW = 1.-sin2thetaW;
    MW         = vev/2.*sqrt((4.*M_PI*alphaQED)/sin2thetaW);
    MZ         = vev/2.*sqrt((4.*M_PI*alphaQED)*(1/sin2thetaW+1/cos2thetaW));
    MH         = p_dataread->GetValue<double>("MH",120.);
    lambdaH    = 2.*sqr(MH/vev);
    break;
  case 3:
    //gmu scheme
    MW         = Flavour(kf_Wplus).Mass();
    MZ         = Flavour(kf_Z).Mass();
    GF         = p_dataread->GetValue<double>("GF",1.16639e-5);
    sin2thetaW = 1.-sqr(MW/MZ);
    cos2thetaW = 1.-sin2thetaW;
    alphaQED   = sqrt(2.)*GF*sqr(MW)*sin2thetaW/M_PI;
    MH         = Flavour(kf_h0).Mass();
    vev        = 1./(pow(2.,0.25)*sqrt(GF));
    lambdaH    = 2.*sqr(MH/vev); 
    break;
  default:
    // all SM parameters given explicitly
    alphaQED   = 1./p_dataread->GetValue<double>("1/ALPHAQED(0)",137.03599976);
    MW         = Flavour(kf_Wplus).Mass();
    MZ         = Flavour(kf_Z).Mass();
    MH         = Flavour(kf_h0).Mass();
    sin2thetaW = p_dataread->GetValue<double>("SIN2THETAW",0.23);
    cos2thetaW = 1.-sin2thetaW;
    vev        = p_dataread->GetValue<double>("VEV",246.);
    lambdaH    = p_dataread->GetValue<double>("LAMBDA",0.47591);
    break;
  }
  aqed                    = new Running_AlphaQED(alphaQED,sqr(MZ));
  double alphaQED_default = 1./p_dataread->GetValue<double>("1/ALPHAQED(default)",1./(*aqed)(sqr(MZ)));
  aqed->SetDefault(alphaQED_default);

  p_functions->insert(std::make_pair(std::string("alpha_QED"),aqed));
  
  if (m_ewscheme!=3) GF = sqrt(2.)*(*aqed)(sqr(Flavour(kf_mu).PSMass()))*M_PI/(2.*sin2thetaW*sqr(MW));
  
  p_constants->insert(std::make_pair(std::string("alpha_QED(0)"),alphaQED));
  p_constants->insert(std::make_pair(std::string("sin2_thetaW"), sin2thetaW));
  p_constants->insert(std::make_pair(std::string("cos2_thetaW"), cos2thetaW));
  p_constants->insert(std::make_pair(std::string("vev"),         vev));
  p_constants->insert(std::make_pair(std::string("MW"),          MW));
  p_constants->insert(std::make_pair(std::string("MZ"),          MZ));
  p_constants->insert(std::make_pair(std::string("GammaW"),      Flavour(kf_Wplus).Width()));
  p_constants->insert(std::make_pair(std::string("GammaZ"),      Flavour(kf_Z).Width()));
  p_constants->insert(std::make_pair(std::string("MH"),          MH));
  p_constants->insert(std::make_pair(std::string("lambdaH"),     lambdaH));
  p_constants->insert(std::make_pair(std::string("GF"),          GF));
}

void Standard_Model::FixCKM() {
  CMatrix CKM(3);

  for (int i=0;i<3;i++) {
    for (int j=i;j<3;j++) CKM[i][j] = CKM[j][i] = Complex(0.,0.);
    CKM[i][i] = Complex(1.,0.);
  }
  
  double Cabibbo=0.0,A=.8,rho,eta;
  m_ckmorder     = p_dataread->GetValue<int>("CKMORDER",0);  
  if (m_ckmorder>0) {
    Cabibbo    = p_dataread->GetValue<double>("CABIBBO",0.2272);
    CKM[0][0] += sqr(Cabibbo)/2. * Complex(-1.,0.);
    CKM[1][1] += sqr(Cabibbo)/2. * Complex(-1.,0.);
    CKM[0][1] += Cabibbo * Complex( 1.,0.);
    CKM[1][0] += Cabibbo * Complex(-1.,0.);
  }
  if (m_ckmorder>1) {
    A          = p_dataread->GetValue<double>("A",0.818);
    CKM[1][2] += A*sqr(Cabibbo)  * Complex( 1.,0.);
    CKM[2][1] += A*sqr(Cabibbo)  * Complex(-1.,0.);
  }
  if (m_ckmorder>2) {
    eta        = p_dataread->GetValue<double>("ETA",0.349);
    rho        = p_dataread->GetValue<double>("RHO",0.227);
    CKM[0][2] += A*pow(Cabibbo,3) * Complex(rho,-eta);
    CKM[2][0] += A*pow(Cabibbo,3) * Complex(1.-rho,-eta);
  }
  p_matrices->insert(std::make_pair(std::string("CKM"),CKM));
}

